"""
Postgres database adaptor.
"""
import asyncio
from asyncio import TimeoutError as aTimeoutError
from contextlib import asynccontextmanager
from time import perf_counter
from typing import Any, AsyncGenerator, Literal, Optional, Tuple, Union

import async_timeout
from asyncpg import Record
from asyncpg.pool import PoolAcquireContext
from sqlalchemy import Sequence, sql
from sqlalchemy.dialects.postgresql import pypostgresql
from sqlalchemy.dialects.postgresql.pypostgresql import PGDialect_pypostgresql
from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.ddl import DDLElement

from ..errors.errors import Error
from ..errors.exception import VLException
from ..utils.log import Logger
from .connection import AbstractDBConnection
from .pool import AbstractSessionPool


def getDialect(**kwargs) -> PGDialect_pypostgresql:
    """
    Get postgresql dialect | http://git.visionlabs.ru/luna/asyncpgsa/-/blob/master/asyncpgsa/connection.py
    """
    dialect = pypostgresql.dialect(paramstyle="pyformat", **kwargs)

    dialect.implicit_returning = True
    dialect.supports_native_enum = True
    dialect.supports_smallserial = True  # 9.2+
    dialect._backslash_escapes = False
    dialect.supports_sane_multi_rowcount = True  # psycopg 2.0.9+
    dialect._has_native_hstore = True

    return dialect


#: postgresql dialect | required to be global to improve requests speed
POSTGRES_DIALECT = getDialect()


class PostgresSessionPool(AbstractSessionPool):
    """
    Session pool for postgres

    Attributes:
        dsn (str): connection URI in the following format: scheme://user:pass@host:port/database?option=value
        dblogin (str): database login
        dbpassword (str): database password
        pool (Pool): connection pool
        connectTimeout (int): connect timeout
    """

    def __init__(self, dsn, dblogin: str, dbpassword: str, connectTimeout: int):
        super().__init__(dsn, dblogin, dbpassword)
        self.connectTimeout = connectTimeout

    async def initialize(self, sessionCount: int, geometryType: Optional[Literal["geometry", "geography"]] = None):
        """
        Initialize connection pool
        Args:
            sessionCount: session count
            geometryType: geometry type  - "geometry"/"geography" or None if no geometry used in database -
                           to initialize connection with custom postgis geometry/geography codec.
                           If geometry type is not specified no geometry codec will be initialized
        """
        from asyncpg import create_pool  # pylint: disable-msg=C0415

        async def initWithGeometry(conn: "asyncpg.connection.Connection"):
            """
            Initialize connection with custom postgis geometry/geography codec
            Args:
                conn: connection
            """

            def encodeGeometry(geometry: "geo_utils.types.Geometry") -> bytes:
                """Encode geometry"""
                import shapely.wkb  # pylint: disable-msg=C0415

                return shapely.wkb.dumps(geometry.shape)

            def decodeGeometry(wkb: bytes) -> "shapely.geometry.base.BaseGeometry":
                """Decode geometry"""
                import shapely.wkb  # pylint: disable-msg=C0415

                return shapely.wkb.loads(wkb)

            schemaReply = await conn.fetchrow(
                f"select nspname from pg_catalog.pg_type c JOIN pg_namespace n ON n.oid = c.typnamespace "
                f"where typname = '{geometryType}'"
            )
            await conn.set_type_codec(
                geometryType, schema=schemaReply[0], encoder=encodeGeometry, decoder=decodeGeometry, format="binary"
            )

        logger = Logger(template="postgres")

        try:
            self.pool = await create_pool(
                dsn=f"postgres://{self.dsn}",
                user=self.dblogin,
                password=self.dbpassword,
                min_size=1,
                max_size=sessionCount,
                init=initWithGeometry if geometryType else None,
                # connect timeout to asyncpg.connection.connect
                timeout=self.connectTimeout,
            )
        except Exception as ex:
            self.printConnectionErrorToLog(ex, logger)
            raise

    @asynccontextmanager
    async def connection(self, logger: Logger) -> AsyncGenerator["asyncpg.transaction.Transaction", None]:
        # pylint: disable-msg=too-many-branches,too-many-statements
        """
        Get connection to postgres
        """
        timeoutBudget = self.connectTimeout
        # not use self.pool.acquire(timeoutBudget) https://github.com/MagicStack/asyncpg/issues/955
        ctx: PoolAcquireContext = self.pool.acquire()
        connection = None
        start = perf_counter()
        try:
            async with async_timeout.timeout(timeoutBudget):
                connection = await ctx.__aenter__()  # pylint: disable-msg=unnecessary-dunder-call
        except aTimeoutError as e:
            if connection:
                # connection valid but pool returns it too late. Exception was raised in timeout
                # __aexit__, return connection to loop
                await ctx.__aexit__()
            logger.error("pg connection pool: run out of connections")
            raise VLException(Error.DataBaseConnectionTimeoutError, 500, False) from e
        except ConnectionRefusedError as e:
            raise VLException(Error.OtherDatabaseError.format("Connection refused"), 500, False) from e
        except asyncio.CancelledError:
            # user canceled request
            if connection:
                await ctx.__aexit__()
            raise

        transaction = connection.transaction()
        # Subtract time spent waiting non active connection from the queue
        # this makes interfaces predictable, meaning that we won't get
        # surprising connectTimeout*2 overall timeout time.
        timeoutBudget -= perf_counter() - start

        try:
            async with async_timeout.timeout(timeoutBudget):
                await transaction.start()
        except aTimeoutError as e:
            logger.error("pg connection pool: start transaction timeout")
            # Timeout occured either when starting transaction/making connection
            # In both cases mark connection as invalid - so it will be aborted and put back
            # in the pool as a "fresh" one.
            connection.terminate()
            raise VLException(Error.DataBaseConnectionTimeoutError, 500, False) from e
        except Exception:
            # wtf happens, mark connection as invalid
            connection.terminate()
            raise
        except asyncio.CancelledError:
            # user cancelled request, connection valid, return to pool
            await ctx.__aexit__()
            raise

        try:
            yield connection
        except Exception:
            await transaction.rollback()
            raise
        else:
            await transaction.commit()
        finally:
            await ctx.__aexit__()

    async def close(self) -> None:
        """
        Attempt to gracefully close all connection to db.
        """
        if self.pool:
            await self.pool.close()


class PostgresDBConnection(AbstractDBConnection):
    """
    Wrapper for a postgres db connection
    """

    def __init__(self, connection: "asyncpg.pool.Pool"):
        """
        Initialize session pool or connection - depends on db type
        Args:
            connection: session pool for postgres
        """
        super().__init__(connection)

    @staticmethod
    def _compile(query: Union[ClauseElement, str]) -> Tuple[str, tuple]:
        """
        Compile statement
        Args:
            query: sa or raw string query
        Returns:
            tuple with compiled query and args
        """
        if isinstance(query, str):
            return query, ()

        compiled = query.compile(dialect=POSTGRES_DIALECT)
        if not isinstance(query, DDLElement):
            compiledParams = sorted(compiled.params.items())
            compiledQuery = compiled.string % {key: f"${idx}" for idx, (key, _) in enumerate(compiledParams, start=1)}
            processors = compiled._bind_processors
            args = [processors[key](val) if key in processors else val for key, val in compiledParams]
        else:
            compiledQuery, args = compiled.string, ()

        return compiledQuery, args

    async def execute(self, statement: sql.ClauseElement, data: Optional[Any] = None) -> int:
        """
        Description see :func:`~AbstractDBConnection.execute`.
        """
        query, params = self._compile(statement)
        commandStatus = await self.connection.execute(query, *params)
        # asyncpg return status of the last SQL command.
        # example: 'INSERT 0 2'
        try:
            return int(commandStatus.split(" ")[-1])
        except ValueError:
            # CREATE TABLE
            return 0

    async def fetchone(self, statement: sql.ClauseElement) -> Union[Tuple, None]:
        """
        Description see :func:`~AbstractDBConnection.fetchone`.
        """
        query, params = self._compile(statement)
        return await self.connection.fetchrow(query, *params)

    async def scalar(self, statement: sql.ClauseElement) -> Union[int, str, float, bytes, None]:
        """
        Description see :func:`~AbstractDBConnection.scalar`.
        """
        query, params = self._compile(statement)
        res = await self.connection.fetchrow(query, *params)
        return res[0] if res is not None else None

    async def nextVal(self, sequence: Sequence) -> int:
        """
        Description see :func:`~AbstractDBConnection.nextVal`.
        """
        val = await self.connection.fetchrow(f"SELECT nextval('{sequence}')")
        return next(val.values())

    async def fetchall(self, statement: sql.ClauseElement) -> list[Record]:
        """
        Description see :func:`~AbstractDBConnection.fetchall`.
        """
        query, params = self._compile(statement)
        rows = await self.connection.fetch(query, *params)
        return rows
