"""Simple process manager with support of configuration reload."""

import asyncio
import os
import platform
import signal
import socket
import sys
from functools import partial
from typing import Awaitable, Callable, Type

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from tzlocal import get_localzone_name

from ..constants.manager import ATTEMPT_RELOAD_COUNT, PROCESS_CHECK_PERIOD
from ..utils.log import Logger
from .config_loader import ConfigLoader, InvalidConfigError, RetryablePuller
from .process import Process
from .reloaders import LoggingReloader

try:
    import coverage
except ImportError:
    coverage = None


class Manager:
    """
    Main process manager class.

    Attributes:
        loop: asyncio event loop
        processes: List of managed processes
        logger: Logger instance
        scheduler: Scheduler for background jobs
        initialConfig: Config from the moment manager has started
        ctxClass: application context class
        getChecks: DEPRECATED
        cov: coveragepy instace. used for tracking coverage during run within pipeline
        sock: socket to pass to child processes
    """

    def __init__(
        self,
        processes: list[Process],
        initialConfig: "BasePythonServiceSettingsClass",
        getChecks: Callable[["BasePythonServiceSettingsClass"], list[Awaitable]] | None = None,
        ctxClass: Type["LunaAppCtx"] | None = None,
        sock: socket.socket | None = None,
    ):
        self.loop = asyncio.get_event_loop()
        if platform.system() != "Windows":
            self.loop.add_signal_handler(signal.SIGTERM, self.handleSignal, signal.SIGTERM)
            self.loop.add_signal_handler(signal.SIGINT, self.handleSignal, signal.SIGINT)
        self.processes = processes
        self.logger = Logger("manager")
        self.scheduler = AsyncIOScheduler(timezone=get_localzone_name())
        self.initialConfig = initialConfig
        self.getChecks = getChecks
        self.ctxClass = ctxClass
        self.sock = sock
        self.cov = None
        if os.getenv("COVERAGE_FILE") and coverage is not None:
            self.logger.warning("Please note that enabling coverage reduce performance")
            self.cov = coverage.Coverage(concurrency="multiprocessing")
            self.cov.start()

    def enableConfigReload(self, configLoader: ConfigLoader, period: float):
        """
        Configure and start configuration reload.

        Args:
            configLoader: load adaptor
            period: A float, seconds to wait between fetching new configuration
        """
        puller = RetryablePuller(
            configLoader,
            initialConfig=self.initialConfig,
            maxInvalidAttempts=ATTEMPT_RELOAD_COUNT,
            connectionValidator=self.validate,
        )

        puller.subscribe(LoggingReloader(self.initialConfig).onConfigUpdate)

        for process in self.processes:
            if not hasattr(process, "getReloader"):
                continue

            reloader = process.getReloader(self.initialConfig)
            puller.subscribe(reloader.onConfigUpdate)

        self.scheduler.add_job(partial(self.pullConfig, puller), trigger="interval", seconds=period)

    async def validate(self, config: "BasePythonServiceSettingsClass") -> bool:
        """Run system checks, returns whether config is OK."""
        if os.getenv("SKIP_CHECK_CONNECTION"):
            self.logger.warning("SKIP_CHECK_CONNECTION var is set, skipping system checks and config validation.")
            return True

        if self.ctxClass:
            ctx = self.ctxClass(config)
            checks = ctx.getChecks()
        else:
            # COMPAT
            checks = self.getChecks(config)

        try:
            res = await asyncio.gather(*checks)
        except Exception:  # pylint: disable-msg=broad-except
            return False
        finally:
            if self.ctxClass:
                await ctx.close()
        return all(res)

    async def pullConfig(self, puller: RetryablePuller):
        """Periodic task, perform configuration pull."""
        try:
            await puller.pull()
        except InvalidConfigError:
            self.logger.error(f"configuration reload has failed after {puller.invalidAttemptsCount} attempt(s)")
            self.stop()
        except Exception:  # pylint: disable-msg=broad-except
            self.logger.exception()
            self.stop()

    def watchProcess(self, process: Process):
        """Periodic task, perform healthcheck."""
        if process.restarting:
            # Process is not dead but restarting, do nothing
            return

        alive = process.isAlive()

        if not alive:
            self.logger.warning(
                f'Process "{process.name}" has died, stopping. with exitcode "{process.process.exitcode}"'
            )
            self.stop()

    def handleSignal(self, sig):
        """Signal handler."""
        self.logger.info(f"Received signal {signal.Signals(sig).name}. Shutting down.")

        if self.cov:
            self.cov.stop()
            self.cov.save()

        self.stop()

    def stop(self):
        """Stop manager."""
        self.loop.stop()  # child processes will be stopped in the finally block

    def run(self):
        """
        Start processes and run forever.

        Prior to starting bind socket and pass it to child processes.
        """

        if not self.loop.run_until_complete(self.validate(self.initialConfig)):
            self.initialConfig.printConfigValues()
            self.stop()
            # keep the old logic when starting the service
            sys.exit(1)

        try:
            for process in self.processes:
                self.logger.info(f"Starting process {process.name}...")
                process.start()
                self.logger.info(f"The process {process.name} has started")
                self.scheduler.add_job(
                    partial(self.watchProcess, process), trigger="interval", seconds=PROCESS_CHECK_PERIOD
                )

            self.scheduler.start()

            try:
                self.logger.info("Manager has started")
                self.loop.run_forever()
            finally:
                self.logger.info("Stopping the manager")
                self.scheduler.shutdown()

                for process in self.processes:
                    self.logger.info(f"Stopping process {process.name}...")
                    process.stop()

                self.loop.close()
        finally:
            self.close()

        self.logger.info("Manager has stopped")

    def close(self):
        """Cleanup."""
        if self.sock:
            self.sock.close()
            self.sock = None
