"""
Module contains pydantic model loaders with some platform-specific features
"""
from typing import Any, Type

from pydantic import ValidationError
from pydantic.main import BaseModel  # pylint: disable-msg=E0611
from vlutils.structures.pydantic import replaceEqualSuffix

from ..errors.errors import Error
from ..errors.exception import VLException

# pydantic model root key
_ROOT_KEY = "__root__"


def getErrorMsgNLocation(exc: ValidationError) -> tuple[str, str]:
    """
    Get error message and location with `discriminated union` model names
    Args:
        exc: pydantic exception
    Returns:
        tuple with string with error location and error message
    Notes:
        - remove after https://github.com/pydantic/pydantic/issues/3947 will be resolved
    """
    error = exc.errors()[0]

    if isinstance(error, list):
        error = error[0]
    excLoc = error.get("loc")
    locTuple = excLoc if isinstance(excLoc, tuple) else (excLoc,)
    if not locTuple:
        return "", error["msg"]
    locations = [str(location) for location in locTuple if "[" not in str(location)]

    errorMsg = error["msg"]
    # avoid additional 'or instance of <Model>' error message from pydantic types parsing
    className = error.get("ctx", {}).get("class_name")
    if error["type"] in ("TypeError", "model_type") and className and errorMsg.endswith(error["ctx"]["class_name"]):
        errorMsg = " ".join(errorMsg.split()[:-4])
        if className in locations:
            locations.remove(className)

    if discriminator := error.get("ctx", {}).get("discriminator"):
        # discriminator value represented in both, snake_case and camelCase like 'objectType' | 'object_type'
        discriminator = discriminator.split(" | ")[-1].replace("'", "")
        match error["type"]:
            case "union_tag_missing":
                locations.append(replaceEqualSuffix(discriminator))
                errorMsg = "Field required"
            case "union_tag_invalid":
                locations.append(replaceEqualSuffix(discriminator))

    return ".".join(filter(lambda x: x != _ROOT_KEY, locations)), errorMsg


def loadDataFromJson(data: dict, model: Type[BaseModel]) -> Any:
    """
    Load data from json with pydantic
    Args:
        data: input data
        model: pydantic model

    Returns:
        initialized object
    Raises:
        VLException(Error.InvalidInputJson.format("Invalid input type."), 400, False) if invalid data type
        VLException(Error.BadInputJson.format(locationStr, errorDict["msg"]), 400, False) if any validation error
    """
    try:
        res = model(**data)
        return res
    except TypeError as exc:
        error = Error.InvalidInputJson.format("Invalid input type.")
        raise VLException(error, 400, isCriticalError=False) from exc
    except (ValidationError, ValueError) as exc:
        errorMsg = getErrorMsgNLocation(exc)
        raise VLException(Error.BadInputJson.format(*errorMsg), 400, isCriticalError=False) from exc


def loadDataFromQuery(data: dict, model: Type[BaseModel]) -> Any:
    """
    Load data from query with pydantic
    Args:
        data: input data
        model: pydantic model

    Returns:
        initialized object
    Raises:
        VLException(Error.BadQueryExtended.format(locationStr, errorDict["msg"]), 400, False) if any validation error
    """
    try:
        res = model(**data)
        return res
    except (ValidationError, TypeError, ValueError) as exc:
        errorDict = exc.errors()[0]
        if errorDict["type"] == "missing":
            raise VLException(Error.QueryParameterNotFound.format(errorDict["loc"][0]), 400, False) from exc
        locationStr = ".".join(filter(lambda x: x != _ROOT_KEY, [str(element) for element in errorDict["loc"]]))
        error = Error.BadQueryParams.format(locationStr)
        raise VLException(error, 400, isCriticalError=False) from exc
