"""
Module contains platform-specific types
"""
import re
from datetime import datetime
from math import isnan

import fastjsonschema
from annotated_types import Ge, Le
from dateutil import tz
from dateutil.relativedelta import relativedelta
from pydantic import Field
from pydantic.types import (  # pylint: disable-msg=no-name-in-module
    confloat as _confloat,
    conint as _conint,
    constr as _constr,
)
from typing_extensions import Annotated
from vlutils.regexps import UUID_REGEXP_STR

from ..errors.pydantic_errors import PydanticError
from ..utils.regexps import VALID_DB_STRING_PATTERN


def OptionalNotNullable() -> object:  # pylint: disable-msg=invalid-name
    """
    Pydantic hack for not required not nullable fields. If field does not present in an input json will be
    initialized of the None, if field is present as null in the input json will be raise corresponding error.
    Examples:
        >>> from pydantic import BaseModel
        >>> class A(BaseModel):
        >>>     id: int
        >>>     description: str = OptionalNotNullable()
        if input json is equal to `{"id": 1, "description": None}` will be raize the error.
        if input json is equal to `{"id": 1}` will be correct parsing.
    Returns:
        Field with default None
    """
    return Field(default_factory=lambda: None)


def Int(minValue: int | None = None, maxValue: int | None = None):  # pylint: disable-msg=invalid-name
    """Get pydantic int model with min/max length"""
    return _conint(ge=minValue, le=maxValue)


def Float(minValue: float | None = None, maxValue: float | None = None):  # pylint: disable-msg=invalid-name
    """Get pydantic float model with min/max length"""
    return _confloat(ge=minValue, le=maxValue)


def Str(minLength: int | None = None, maxLength: int | None = None):  # pylint: disable-msg=invalid-name
    """Get pydantic strict string model with min/max length"""
    return _constr(strict=True, pattern=VALID_DB_STRING_PATTERN, min_length=minLength, max_length=maxLength)


def QueryStr(minLength: int | None = None, maxLength: int | None = None):  # pylint: disable-msg=invalid-name
    """Get pydantic strict string model with min/max length"""
    return _constr(strict=False, pattern=VALID_DB_STRING_PATTERN, min_length=minLength, max_length=maxLength)


class _CustomFloatMeta(type):
    """Custom float meta"""

    def __getitem__(cls, param: any) -> type:
        return type("CustomFloat", (CustomFloat,), {"param": param})


ObjectId = _constr(strict=True, pattern=rf"^{UUID_REGEXP_STR}")
TrackId = _constr(strict=True, pattern=r"^[a-zA-Z0-9_\-]{1,36}$")

DATETIME_SCHEMA = fastjsonschema.compile({"type": "string", "format": "date-time"})
DATETIME_NOW_SCHEMA = fastjsonschema.compile({"type": "string", "pattern": r"^now-(\d+)[smhdwMy]$"})

Latitude = Annotated[float, Ge(-90.0), Le(90.0)]
Longitude = Annotated[float, Ge(-180.0), Le(180.0)]


class _CustomIntMeta(type):
    """Custom float meta"""

    def __getitem__(cls, param: any) -> type:
        return type("CustomInt", (CustomInt,), {"param": param})


class _CustomStringMeta(type):
    """Custom float meta"""

    def __getitem__(cls, param: any) -> type:
        return type("CustomString", (CustomString,), {"param": param})


class CustomNumberMixin:
    """Mixin for Custom number types that provides common properties and validators"""

    gt: int | float | None = None
    ge: int | float | None = None
    lt: int | float | None = None
    le: int | float | None = None
    param: any = None
    multiple_of: int | float | None = None
    strict: bool = True

    @classmethod
    def validateNonStr(cls, value: float | int, *args, **kwargs) -> float:  # pylint: disable-msg=unused-argument
        """Validate value is not string"""
        if isinstance(value, str):
            raise PydanticError.PydanticValueNotANumberError()
        return value

    @classmethod
    def intValidator(cls, value: int, *args, **kwargs):  # pylint: disable-msg=unused-argument
        """
        Validate value is an integer or might be converted to it

        Args:
            value: object
        Returns:
            int
        Raises:
            TypeError: if value can not be converted to int type
        """
        if isinstance(value, int):
            return value
        raise PydanticError.PydanticValueNotAnIntegerError()

    @classmethod
    def floatValidator(cls, value: float, *args, **kwargs) -> float:  # pylint: disable-msg=unused-argument
        """
        Validate value is float or might be converted to it

        Args:
            value: object
        Returns:
            float
        Raises:
            TypeError: if value can not be converted to float type
        """

        if isinstance(value, float):
            return value
        try:
            return float(value)
        except (TypeError, ValueError) as exc:
            raise PydanticError.PydanticValueNotAFloatError() from exc

    @classmethod
    def numberSizeValidator(  # pylint: disable-msg=unused-argument
        cls, value: float | int, *args, **kwargs
    ) -> float:  # pylint: disable-msg=unused-argument
        """
        Validate that number respects the limits

        Args:
            value: float | int
        Returns:
            float | int
        Raises:
            ValidationError: if the specified limits are exceed
        """

        if cls.gt is not None and value <= cls.gt:
            raise PydanticError.PydanticValidationError.format(f"Value must be greater than {cls.gt}").setContext(
                {"limit_value": cls.gt}
            )()
        if cls.ge is not None and value < cls.ge:
            raise PydanticError.PydanticValidationError.format(
                f"Value must be greater than or equal {cls.ge}"
            ).setContext({"limit_value": cls.ge})()

        if cls.lt is not None and value >= cls.lt:
            raise PydanticError.PydanticValidationError.format(f"Value must be less than {cls.lt}").setContext(
                {"limit_value": cls.lt}
            )()
        if cls.le is not None and value > cls.le:
            raise PydanticError.PydanticValidationError.format(f"Value must be less than or equal {cls.le}").setContext(
                {"limit_value": cls.le}
            )()

        return value

    @classmethod
    def numberMultipleValidator(
        cls, value: float | int, *args, **kwargs  # pylint: disable-msg=unused-argument
    ) -> float:
        """
        Ensure that the number is multiple of the specified multiple_of property

        Args:
            value: float | int

        Returns:
            float | int
        Raises:
            ValidationError: if the specified multiple_of parameter is not correct
        """
        if cls.multiple_of is not None:
            mod = float(value) / float(cls.multiple_of) % 1
            delta = 1e-8
            if abs(mod - 0.0) <= delta or abs(mod - 1.0) <= delta:
                raise PydanticError.PydanticValidationError.format(
                    f"Ensure this value is a multiple of {cls.multiple_of}"
                ).setContext(
                    {"multiple_of": cls.multiple_of},
                )()
        return value


class CustomInt(int, CustomNumberMixin, metaclass=_CustomIntMeta):
    """
    Custom int - same as default pydantic non-strict int, but not allow to use string as int
    """

    @classmethod
    def __get_validators__(cls):
        """Use own validator."""
        if cls.strict:
            yield cls.validateNonStr
        yield cls.intValidator
        yield cls.numberSizeValidator
        yield cls.numberMultipleValidator


class CustomFloat(float, CustomNumberMixin, metaclass=_CustomFloatMeta):
    """
    Custom float - same as default pydantic non-strict float, but not allow to use string as float
    """

    forbidNaN = False

    @classmethod
    def __get_validators__(cls):
        """Use own validator."""
        if cls.strict:
            yield cls.validateNonStr
        yield cls.floatValidator
        yield cls.numberSizeValidator
        yield cls.numberMultipleValidator
        if cls.forbidNaN:
            yield cls.validateNotNaN

    @classmethod
    def validateNotNaN(cls, value: float, *args, **kwargs) -> float:  # pylint: disable-msg=unused-argument
        """Validate that value is not NaN"""
        if isnan(value):
            raise PydanticError.PydanticValueNotANumberError()
        return value


def validateUniqueItems(value: list, *args, **kwargs) -> list:  # pylint: disable-msg=unused-argument
    """Validate that all the items in list are unique"""
    if len(value) != len(set(value)):
        raise PydanticError.PydanticUniqueItemsError()
    return value


class CustomString(str, metaclass=_CustomStringMeta):
    """Custom string type"""

    min_length: int | None = None
    max_length: int | None = None
    pattern: str | None = None

    @classmethod
    def __get_validators__(cls):
        """Use own validator."""
        yield cls.stringValidator
        yield cls.lengthValidator
        if cls.pattern:
            yield cls.patternValidator

    @classmethod
    def stringValidator(cls, value: str, *args, **kwargs) -> str:  # pylint: disable-msg=unused-argument
        """Validate value is a string"""
        if not isinstance(value, str):
            raise PydanticError.PydanticInvalidStringError()
        return value

    @classmethod
    def patternValidator(cls, value: str, *args, **kwargs) -> str:  # pylint: disable-msg=unused-argument
        """
        Validates that string matches the pattern

        Raises PydanticCustomError("ValueError", f'String Pattern mismatch - "<pattern>"')
        if string does not match pattern
        """

        if not re.match(cls.pattern, value):
            raise PydanticError.PydanticStringPatternMismatchError.format(cls.pattern)()
        return value

    @classmethod
    def lengthValidator(cls, value: str, *args, **kwargs) -> str:  # pylint: disable-msg=unused-argument
        """
        Validates that string length respects the set limits

        Raises ValidationError('Value length must be less than <max_length>') if length is greater than max_length
        Raises ValidationError('Value length must be less than <min_length>') if length is less than min_length
        """

        if cls.min_length is not None and len(value) < cls.min_length:
            raise PydanticError.PydanticValidationError.format(
                f"Value length must be greater than {cls.min_length}"
            ).setContext({"limit_value": cls.min_length})()
        if cls.max_length is not None and len(value) > cls.max_length:
            raise PydanticError.PydanticValidationError.format(
                f"Value length must be less than {cls.max_length}"
            ).setContext({"limit_value": cls.max_length})()
        return value


class BoolFromInt01(CustomString):
    """Bool from 0/1 int getter"""

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, *args, **kwargs) -> bool:  # pylint: disable-msg=unused-argument
        """Validate value and morph to bool"""
        try:
            value = int(value)
        except (ValueError, TypeError) as exc:
            raise PydanticError.PydanticValueNotANumberError() from exc
        if value not in (0, 1):
            raise PydanticError.PydanticValidationError.format(f"Value should be either 0 or 1, not {value}")()
        return bool(value)


class _CustomDatetimeMeta(type):
    """Custom datetime meta"""

    def __getitem__(cls, param: any) -> type:
        return type("CustomDatetime", (CustomDatetime,), {"param": param})


class CustomDatetime(metaclass=_CustomDatetimeMeta):
    """
    Custom extended datetime - do not use default if it needs to get attribute as str in ISO-format,
    """

    @classmethod
    def __get_validators__(cls):
        """Use own validator."""
        yield cls.validateDatetime

    @classmethod
    def validateDatetime(cls, value: str | datetime, *args, **kwargs) -> str:  # pylint: disable-msg=unused-argument
        """
        Validate value is not string
        Args:
            value: input value
        Returns:
            datetime in string format
        """
        try:
            DATETIME_SCHEMA(value)
            return str(value)
        except fastjsonschema.exceptions.JsonSchemaException:  # pylint: disable-msg=W0612
            raise PydanticError.PydanticInvalidDateTimeError() from None


class CustomExtendedDatetime(metaclass=_CustomDatetimeMeta):
    """
    Custom datetime - do not use default if it needs to get attribute as str in ISO-format
    """

    _storageTime: str

    @classmethod
    def initialize(cls, storageTime: str):
        """
        Initialize storage time
        Args:
            storageTime: storage time - utc or local
        """
        cls._storageTime = storageTime

    @classmethod
    def __get_validators__(cls):
        """
        Get validators - use pydantic default validation, but return attribute as valid datetime str in ISO-format
        """
        yield cls.validateDatetime

    @classmethod
    def validateDatetime(cls, value: str | datetime, *args, **kwargs) -> str:  # pylint: disable-msg=unused-argument
        """
        Validate value is not string
        Args:
            value: input value
        Returns:
            datetime in string format
        """
        if isinstance(value, str) and value.startswith("now-"):
            try:
                DATETIME_NOW_SCHEMA(value)
            except fastjsonschema.exceptions.JsonSchemaException:
                raise PydanticError.PydanticInvalidDateTimeError() from None
            splitDelta = value.split("-")[1]
            deltaValue, deltaType = int(splitDelta[:-1]), splitDelta[-1]
            kwarg = dict(y="years", M="months", w="weeks", d="days", h="hours", m="minutes", s="seconds")[deltaType]
            try:
                datetimeResult = (
                    (datetime.utcnow() - relativedelta(**{kwarg: deltaValue})).isoformat() + "Z"
                    if cls._storageTime == "UTC"
                    else (
                        datetime.now().replace(tzinfo=tz.tzlocal()) - relativedelta(**{kwarg: deltaValue})
                    ).isoformat()
                )
            except ValueError:
                raise PydanticError.PydanticInvalidDateTimeError() from None
            return datetimeResult
        try:
            DATETIME_SCHEMA(value)
            return str(value)
        except fastjsonschema.exceptions.JsonSchemaException:
            raise PydanticError.PydanticInvalidDateTimeError() from None


class ListStringsValidator(CustomString):
    """List string validator."""

    pattern = VALID_DB_STRING_PATTERN

    @classmethod
    def __get_validators__(cls):
        super().__get_validators__()
        yield cls.splitStringByComas

    @classmethod
    def splitStringByComas(cls, value: str, *args, **kwargs) -> list[str]:  # pylint: disable-msg=unused-argument
        """Create list with values from string."""

        return value.split(",")


def customfloat(*, gt: int = None, ge: int = None, lt: int = None, le: int = None) -> type[float]:
    """Construct Field for custom float."""
    return type("CustomFloatValue", (CustomFloat,), {ge: ge, lt: lt, le: le, gt: gt})
