"""
Base adaptor.
"""
from enum import Enum

from redis.asyncio.client import Redis
from redis.asyncio.cluster import RedisCluster
from redis.asyncio.connection import BlockingConnectionPool
from redis.asyncio.sentinel import Sentinel


class RedisConfiguration(Enum):
    """Redis configuration."""

    SERVER = "server"
    SENTINEL = "sentinel"
    CLUSTER = "cluster"


class RedisAdaptor:
    """
    Base adaptor for Redis DB

    Attributes:
        client (Redis): factory for connection to database
        configuration (RedisConfiguration): database configuration
    """

    def __init__(
        self,
        host: str = None,
        port: int = None,
        username: str = None,
        password: str = None,
        dbNumber: int = None,
        sentinels: list[tuple[str, int]] = None,
        sentinelUsername: str = None,
        sentinelPassword: str = None,
        masterName: str = None,
        clusterDiscovery: bool = None,
        maxConnections: int = None,
    ):
        if sentinels:
            self.configuration = RedisConfiguration.SENTINEL
            sentinel = Sentinel(
                sentinels=sentinels,
                sentinel_kwargs={"username": sentinelUsername, "password": sentinelPassword},
                username=username,
                password=password,
                db=dbNumber,
            )
            if maxConnections:
                pool = BlockingConnectionPool(max_connections=maxConnections)
                self.client = sentinel.master_for(masterName, connection_pool=pool)
            else:
                self.client = sentinel.master_for(masterName)
        elif clusterDiscovery:
            self.configuration = RedisConfiguration.CLUSTER
            connectionKwargs = {"host": host, "port": port, "username": username, "password": password}
            if maxConnections:
                # async RedisCluster has no support for blocking connection pool
                # https://github.com/redis/redis-py/issues/2350
                self.client = RedisCluster(**connectionKwargs, max_connections=maxConnections)
            else:
                self.client = RedisCluster(**connectionKwargs)
        else:
            self.configuration = RedisConfiguration.SERVER
            connectionKwargs = {"host": host, "port": port, "username": username, "password": password, "db": dbNumber}
            if maxConnections:
                pool = BlockingConnectionPool(max_connections=maxConnections, **connectionKwargs)
                self.client = Redis(connection_pool=pool)
            else:
                self.client = Redis(**connectionKwargs)

    async def close(self):
        """Closes client connection."""
        await self.client.close()
