"""
Module contains standard command line arguments parser for application
"""
from argparse import ArgumentParser
from dataclasses import dataclass
from typing import Dict, Optional, Set

from ..constants.paths import CONFIG_FILE


@dataclass
class AppCMDOptions:
    """
    Application cmd options
    """

    # listener port
    port: int
    # worker count
    workers: int
    # config path
    config: str
    # luna-configurator address
    configuratorAddress: Optional[str]
    # log suffix
    logSuffix: Optional[str]
    # setting names and tags for pulling settings from configurator
    tags: Optional[Dict[str, str]]


class ApplicationCMDParser:  # pylint: disable-msg=too-few-public-methods
    """
    Command line parser

    Attributes:
        argParser (ArgumentParser): argument parser
        tagableSettings (Set[str]): set of setting names which allowed tags for pulling
        defaultConfigFilePath (str): default config path
    """

    def __init__(
        self,
        defaultPort: Optional[int] = None,
        defaultConfigFilePath: Optional[str] = None,
        tagableSettings: Optional[Set[str]] = None,
    ):
        self.tagableSettings = tagableSettings
        self.defaultConfigFilePath = defaultConfigFilePath

        argParser = ArgumentParser(add_help=False)
        if defaultPort is not None:
            msg = f"listener port, default {defaultPort}"
        else:
            msg = "listener port"
        argParser.add_argument("--port", default=defaultPort, type=notNegativeInt, help=msg)
        argParser.add_argument("--workers", default=1, type=int, help="worker count")
        argParser.add_argument("--log_suffix", default=None, type=str, help="suffix for log files")
        if defaultConfigFilePath is None:
            defaultConfigFilePath = str(CONFIG_FILE)
        argParser.add_argument("--config", default=str(defaultConfigFilePath), type=str, help="path to config")
        argParser.add_argument(
            "--luna-config",
            default=None,
            type=str,
            help="origin and api version of luna-configurator service for pulling setting from it",
        )
        if tagableSettings:
            for setting in tagableSettings:
                argParser.add_argument(
                    f"--{setting}",
                    default=None,
                    type=str,
                    help=f"tag for setting '{setting}' to pull from luna-configurator",
                )

        self.argParser = argParser

    def parse(self) -> AppCMDOptions:
        """
        Parse command line
        Returns:
            parsed options
        """
        args = self.argParser.parse_known_args()[0]
        settingsWithTags = {
            settingName: args.__dict__[settingName]
            for settingName in self.tagableSettings
            if settingName in args.__dict__ and args.__dict__[settingName] is not None
        }
        return AppCMDOptions(
            port=args.port,
            workers=args.workers,
            logSuffix=args.log_suffix,
            configuratorAddress=args.luna_config,
            tags=settingsWithTags,
            config=args.config,
        )


def notNegativeInt(value: str) -> int:
    """Convert input string to not negative integer, raise TypeError in case input is invalid."""
    ivalue = int(value)
    if ivalue < 0:
        raise TypeError(f"{value} is an invalid positive int value")
    return ivalue
