"""
Context for application.
Defines components and how do they depend on each other.
"""

import asyncio
from argparse import Namespace
from typing import Awaitable, Generic, TypeVar

from luna_plugins.base.manager import PluginManager

from ..monitoring.manager import LunaMonitoringManager
from ..utils import mixins
from ..utils.log import logger

CONFIG_CLASS = TypeVar("CONFIG_CLASS")


class LunaAppCtx(Generic[CONFIG_CLASS]):
    """
    Base luna application context
    Attributes:
        cmdArgs: command line arguments
        logger: global app logger
        serviceConfig: service config
        pluginManager: plugin manager
        monitoring: service monitoring
        apiVersion: current api version
    """

    def __init__(self, config: CONFIG_CLASS, pluginManager: PluginManager | None = None):
        self.logger = logger
        self.serviceConfig: CONFIG_CLASS = config
        self.monitoring = LunaMonitoringManager(settings=config.monitoring, pluginManager=pluginManager)
        self.pluginManager = pluginManager
        self.cmdArgs: Namespace | None = None

    async def initialize(self):
        """
        Initialize service
        """
        coros = [x.initializeWrapper() for x in self if isinstance(x, mixins.Initializable)]

        await asyncio.gather(*coros)

    async def close(self):
        """
        Stop context. Make sure to close components in correct order.
        """
        # reverse so that components with dependencies will go first
        # doesn't guarantee safety
        for x in reversed(list(self)):
            if isinstance(x, mixins.Initializable):
                await x.close()

    def __iter__(self):
        """Yields class attributes for convenience."""
        for value in self.__dict__.values():
            yield value

    def getChecks(self):
        """Returns coroutines to verify configuration."""
        checks = []
        for component in self:
            if hasattr(component, "probe"):
                checks.append(component.probe())

        return checks

    def getRuntimeChecks(self, includeLunaServices: bool) -> list[tuple[str, Awaitable]]:
        """Returns list of coroutines with codes that are used for /healthcheck resouce."""
        checks = []
        for component in self:
            if hasattr(component, "getRuntimeChecks"):
                checks.extend(component.getRuntimeChecks(includeLunaServices))
        return checks


CONTEXT_CLASS = TypeVar("CONTEXT_CLASS", bound=LunaAppCtx)
